from npc_utils import *

from npc_sessions.utils.cache import *
from npc_sessions.utils.electrodes import *
from npc_sessions.utils.intervals import *
from npc_sessions.utils.misc import *
from npc_sessions.utils.newscale import *
from npc_sessions.utils.running import *
from npc_sessions.utils.videos import *



from __future__ import annotations

import importlib.resources
import logging
import os

import npc_session
import upath

import npc_sessions.utils as utils

logger = logging.getLogger(__name__)

MODULE_ROOT = upath.UPath(__file__).parent
PACKAGE_ROOT = utils.from_pathlike(importlib.resources.files("npc_sessions"))  # type: ignore[arg-type]


def write_qc_notebook(
    session_path_or_id: str,
    save_path: utils.PathLike = upath.UPath.cwd(),
    **session_kwargs,
) -> upath.UPath:
    """Find the given session's appropriate QC notebook in the site-packages
    folder, run it and save it at the given path.

    - currently assumes dynamic routing sessions

    >>> output_path = write_qc_notebook("664566_20230403", is_ephys=False)
    """
    QC_NOTEBOOK = next(MODULE_ROOT.rglob("dynamic*routing*qc*.ipynb"))

    # pass config to run in notebook via env vars
    env = {}
    env["NPC_SESSION_PATH_OR_ID"] = session_path_or_id
    for k, v in session_kwargs.items():
        env[f"NPC_SESSION_{k}"] = str(v)
        # notebook should eval or json.loads env vars to pass to session init
        # keys are auto-capitalized (on windows, at least), so also apply .lower()

    logger.info(
        f"running {QC_NOTEBOOK.name} for {npc_session.SessionRecord(session_path_or_id)}"
    )
    return utils.run_and_save_notebook(
        notebook_path=QC_NOTEBOOK,
        save_path=save_path,
        env=dict(os.environ) | env,  # merge with current env to reuse credentials etc.
    )


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



"""
getting trials tables for experiments based on sam's TaskControl requires
timing information about each 'frame' in the experiment event loop, plus any
metadata about trials/stimuli.

Files required:

- a single hdf5 file, incl trial/stim data and frame display times (not
  synced with other devices)

- [optional for single hdf5] vsync times on sync clock, which give more precise, synced timing
  of each 'frame' of the experiment, used in place of the times in the hdf5

- [optional] for aud, opto, other non-visual stimuli: latency estimates from
  synced recordings of stimulus presentations, added to vsync times

Timing info will be obtained from the first available source from the left:

custom overrides (e.g non-vis stim latencies) > vsync times > hdf5 times

Info for custom overrides should be passed as kwargs to the constructor, e.g.:
TaskControl('stim.hdf5', sync='sync.h5', aud_stim_latencies=np.array([np.nan, 0.01, np.nan]))

`aud_stim_latencies` will be added as an instance attribute with a leading
underscore, and can then used where needed in the class, e.g.:
@property
def stim_start_times(self):
    return self._stim_onsets_from_vsyncs + self._aud_stim_latencies

# Concatenating trials
Where there are multiple sets of trials for the same experiment (e.g.
optotagging pre- and post-task) it should be possible to concatentate them
by just adding trials to an existing nwb table, via `nwb_file.add_trial(**trial)`
 - just be careful that shared parameters are referring to the same across all
   concatenated trials, e.g.  stim_idx 1 in both sets must be the same stim
 - any `trial_idx` type counters might need updating too
"""

from __future__ import annotations

import functools
from collections.abc import Iterable
from typing import Optional, SupportsFloat

import h5py
import numpy as np
import numpy.typing as npt

import npc_sessions.trials.property_dict as property_dict
import npc_sessions.utils as utils


class TaskControl(property_dict.PropertyDict):
    _sync: utils.SyncDataset | None
    _hdf5: h5py.File

    def __init__(
        self,
        hdf5: utils.StimPathOrDataset,
        sync: utils.SyncPathOrDataset | None = None,
        **kwargs,
    ) -> None:
        for k, v in kwargs.items():
            setattr(self, f"_{k.strip('_')}", v)
        self._hdf5 = utils.get_h5_stim_data(hdf5)
        if not sync:
            # - all times in nwb are relative to start of first frame in hdf5
            # - there can only be one hdf5 file
            self._sync = None
        else:
            # - all times in nwb are relative to start of first sample on sync
            # - there can be multiple hdf5 files, all recorded on sync
            self._sync = utils.get_sync_data(sync)

    @utils.cached_property
    def _input_data_times(self) -> npt.NDArray[np.float64]:
        """Best-estimate time of `getInputData()` in psychopy event loop, in seconds, from start
        of experiment. Uses preceding frame's vsync time if available."""
        return utils.get_input_data_times(self._hdf5, self._sync)

    @utils.cached_property
    def _flip_times(self) -> npt.NDArray[np.float64]:
        """Best-estimate time of `flip()` in psychopy event loop, in seconds, from start
        of experiment. Uses frame's vsync time if available."""
        return utils.get_flip_times(self._hdf5, self._sync)

    @utils.cached_property
    def _vis_display_times(self) -> npt.NDArray[np.float64]:
        """Best-estimate time of monitor update. Without sync, this equals frame times."""
        return utils.get_vis_display_times(self._hdf5, self._sync)



from __future__ import annotations

import concurrent.futures
import datetime
import logging
import logging.config
import multiprocessing
import time
from typing import Literal

import npc_lims
import npc_session
import tqdm

import npc_sessions
import npc_sessions.scripts.utils as utils

logger = logging.getLogger()


def helper(
    session_id: str | npc_session.SessionRecord | npc_lims.SessionInfo,
    **write_all_components_kwargs,
) -> None:
    session = npc_sessions.DynamicRoutingSession(session_id)
    logger.info(f"Processing {session.id}")
    npc_sessions.write_all_components_to_cache(
        session,
        **write_all_components_kwargs,
    )
    del session


def write_sessions_to_cache(
    session_type: Literal["training", "ephys", "all"] = "all",
    skip_existing: bool = True,
    version: str | None = None,
    parallel: bool = True,
    max_workers: int | None = None,
) -> None:
    t0 = time.time()
    session_infos = utils.get_session_infos(session_type=session_type)

    helper_opts = {
        "version": version,
        "skip_existing": skip_existing,
    }
    if len(session_infos) == 1:
        parallel = False
    if parallel:
        future_to_session = {}
        pool = concurrent.futures.ProcessPoolExecutor(
            max_workers or utils.get_max_workers(session_type),
            mp_context=multiprocessing.get_context("spawn"),
        )
        for info in tqdm.tqdm(session_infos, desc="Submitting jobs"):
            future_to_session[
                pool.submit(
                    helper,
                    session_id=info,
                    **helper_opts,  # type: ignore[arg-type]
                )
            ] = info.id
        for future in tqdm.tqdm(
            concurrent.futures.as_completed(future_to_session.keys()),
            desc="Processing jobs",
            total=len(future_to_session),
        ):
            logger.info(f"{future_to_session[future]} done")
            continue
    else:
        for info in tqdm.tqdm(session_infos, desc="Processing jobs"):
            helper(
                session_id=info,
                **helper_opts,  # type: ignore[arg-type]
            )
            logger.info(f"{info.id} done")
    npc_sessions.consolidate_all_caches()
    logger.info(f"Time elapsed: {datetime.timedelta(seconds=time.time() - t0)}")


def main() -> None:
    kwargs = utils.setup()
    write_sessions_to_cache(**kwargs)  # type: ignore[misc, arg-type]


if __name__ == "__main__":
    main()



from __future__ import annotations

import tempfile

import ndx_pose
import npc_lims
import numpy as np
import numpy.typing as npt
import pandas as pd
import upath
from scipy import ndimage, stats

# nice little trick from carter peene
MODEL_FUNCTION_MAPPING = {
    "dlc_eye": npc_lims.get_dlc_eye_s3_paths,
    "dlc_side": npc_lims.get_dlc_side_s3_paths,
    "dlc_face": npc_lims.get_dlc_face_s3_paths,
}


def get_dlc_session_paf_graph(session: str, model_name: str) -> list:
    """
    https://github.com/DeepLabCut/DLC2NWB/blob/main/dlc2nwb/utils.py#L139
    >>> get_dlc_session_paf_graph('676909_2023-12-12', 'dlc_eye')
    []
    """
    model_s3_paths = MODEL_FUNCTION_MAPPING[model_name](session)
    metadata_pickle_file = tuple(
        path for path in model_s3_paths if path.suffix == ".pickle"
    )
    if not metadata_pickle_file:
        raise FileNotFoundError(
            f"No metadata pickle found for session {session}. Check {model_name} capsule"
        )

    metadata_pickle = pd.read_pickle(metadata_pickle_file[0])
    test_config = metadata_pickle["data"]["DLC-model-config file"]
    paf_graph = test_config.get("partaffinityfield_graph", [])
    if paf_graph:
        paf_inds = test_config.get("paf_best")
        if paf_inds is not None:
            paf_graph = [paf_graph[i] for i in paf_inds]

    return paf_graph


def h5_to_dataframe(h5_path: upath.UPath, key_name: str | None = None) -> pd.DataFrame:
    with tempfile.TemporaryDirectory() as tempdir:
        with open(f"{tempdir}/{h5_path.stem}", "wb") as h5_file:
            h5_file.write(h5_path.read_bytes())
            if key_name is not None:
                df_h5 = pd.read_hdf(f"{tempdir}/{h5_path.stem}", key=key_name)
            else:
                df_h5 = pd.read_hdf(f"{tempdir}/{h5_path.stem}")

    return df_h5


def compute_elliptical_area(df_row: pd.Series) -> float:
    """Calculate the area of corneal reflection (cr) or eye ellipse fits using
    the ellipse formula.

    Parameters
    ----------
    df_row : pd.Series
        A row from an eye tracking dataframe containing either:
        "cr_width", "cr_height"
        or
        "eye_width", "eye_height"

    Returns
    -------
    float
        The elliptical area of the eye or cr in pixels^2
    """
    return np.pi * df_row.iloc[0] * df_row.iloc[1]


def compute_circular_area(df_row: pd.Series) -> float:
    """Calculate the area of the pupil as a circle using the max of the
    height/width as radius.

    Note: This calculation assumes that the pupil is a perfect circle
    and any eccentricity is a result of the angle at which the pupil is
    being viewed.

    Parameters
    ----------
    df_row : pd.Series
        A row from an eye tracking dataframe containing only "pupil_width"
        and "pupil_height".

    Returns
    -------
    float
        The circular area of the pupil in pixels^2.
    """
    max_dim = max(df_row.iloc[0], df_row.iloc[1])
    return np.pi * max_dim * max_dim


def determine_outliers(data_df: pd.DataFrame, z_threshold: float) -> pd.Series:
    """Given a dataframe and some z-score threshold return a pandas boolean
    Series where each entry indicates whether a given row contains at least
    one outlier (where outliers are calculated along columns).

    Parameters
    ----------
    data_df : pd.DataFrame
        A dataframe containing only columns where outlier detection is
        desired. (e.g. "cr_area", "eye_area", "pupil_area")
    z_threshold : float
        z-score values higher than the z_threshold will be considered outliers.

    Returns
    -------
    pd.Series
        A pandas boolean Series whose length == len(data_df.index).
        True denotes that a row in the data_df contains at least one outlier.
    """

    outliers = (
        data_df.apply(stats.zscore, nan_policy="omit").apply(np.abs) > z_threshold
    )
    return pd.Series(outliers.any(axis=1))


def determine_likely_blinks(
    eye_areas: pd.Series,
    pupil_areas: pd.Series,
    outliers: pd.Series,
    dilation_frames: int = 2,
) -> pd.Series:
    """Determine eye tracking frames which contain likely blinks or outliers

    Parameters
    ----------
    eye_areas : pd.Series
        A pandas series of eye areas.
    pupil_areas : pd.Series
        A pandas series of pupil areas.
    outliers : pd.Series
        A pandas series containing bool values of outlier rows.
    dilation_frames : int, optional
        Determines the number of additional adjacent frames to mark as
        'likely_blink', by default 2.

    Returns
    -------
    pd.Series
        A pandas series of bool values that has the same length as the number
        of eye tracking dataframe rows (frames).
    """
    blinks = pd.isnull(eye_areas) | pd.isnull(pupil_areas) | outliers
    if dilation_frames > 0:
        likely_blinks = ndimage.binary_dilation(blinks, iterations=dilation_frames)
    else:
        likely_blinks = blinks
    return pd.Series(likely_blinks, index=eye_areas.index)


def filter_on_blinks(eye_tracking_data: pd.DataFrame):
    """Set data is specified columns where likely_blink is true to NaN.

    Modify the DataFrame in place.

    Parameters
    ----------
    eye_tracking_data : pandas.DataFrame
        Data frame containing eye tracking data.
    """
    likely_blinks = eye_tracking_data["likely_blink"]
    eye_tracking_data.loc[likely_blinks, "eye_area"] = np.nan
    eye_tracking_data.loc[likely_blinks, "pupil_area"] = np.nan
    eye_tracking_data.loc[likely_blinks, "cr_area"] = np.nan

    eye_tracking_data.loc[likely_blinks, "eye_width"] = np.nan
    eye_tracking_data.loc[likely_blinks, "eye_height"] = np.nan
    eye_tracking_data.loc[likely_blinks, "eye_phi"] = np.nan

    eye_tracking_data.loc[likely_blinks, "pupil_width"] = np.nan
    eye_tracking_data.loc[likely_blinks, "pupil_height"] = np.nan
    eye_tracking_data.loc[likely_blinks, "pupil_phi"] = np.nan


def get_computed_ellipse_metrics_dataframe(
    eye_data: pd.DataFrame, z_threshold: float = 3.0, dilation_frames: int = 2
) -> pd.DataFrame:
    """
    >>> df_ellipse = get_ellipse_session_dataframe_from_h5('676909_2023-12-12')
    >>> df_eliipse_computed_metrics = get_computed_ellipse_metrics_dataframe(df_ellipse)
    >>> df_eliipse_computed_metrics.columns
    Index(['cr_area', 'eye_area', 'pupil_area', 'likely_blink', 'pupil_area_raw',
           'cr_area_raw', 'eye_area_raw', 'cr_center_x', 'cr_center_y', 'cr_width',
           'cr_height', 'cr_phi', 'eye_center_x', 'eye_center_y', 'eye_width',
           'eye_height', 'eye_phi', 'pupil_center_x', 'pupil_center_y',
           'pupil_width', 'pupil_height', 'pupil_phi'],
          dtype='object')
    """
    cr_areas = eye_data[["cr_width", "cr_height"]].apply(
        compute_elliptical_area, axis=1
    )
    eye_areas = eye_data[["eye_width", "eye_height"]].apply(
        compute_elliptical_area, axis=1
    )
    pupil_areas = eye_data[["pupil_width", "pupil_height"]].apply(
        compute_circular_area, axis=1
    )

    # only use eye and pupil areas for outlier detection
    area_df = pd.concat([eye_areas, pupil_areas], axis=1)
    outliers = determine_outliers(area_df, z_threshold=z_threshold)

    likely_blinks = determine_likely_blinks(
        eye_areas, pupil_areas, outliers, dilation_frames=dilation_frames
    )

    # remove outliers/likely blinks `pupil_area`, `cr_area`, `eye_area`
    pupil_areas_raw = pupil_areas.copy()
    cr_areas_raw = cr_areas.copy()
    eye_areas_raw = eye_areas.copy()

    eye_data.insert(0, "cr_area", cr_areas)
    eye_data.insert(1, "eye_area", eye_areas)
    eye_data.insert(2, "pupil_area", pupil_areas)
    eye_data.insert(3, "likely_blink", likely_blinks)
    eye_data.insert(4, "pupil_area_raw", pupil_areas_raw)
    eye_data.insert(5, "cr_area_raw", cr_areas_raw)
    eye_data.insert(6, "eye_area_raw", eye_areas_raw)

    # Apply blink fliter to additional columns.
    filter_on_blinks(eye_data)

    return eye_data


def get_ellipse_session_dataframe_from_h5(session: str) -> pd.DataFrame:
    """
    >>> df_ellipse = get_ellipse_session_dataframe_from_h5('676909_2023-12-12')
    >>> len(df_ellipse)
    512347
    """
    eye_s3_paths = npc_lims.get_dlc_eye_s3_paths(session)
    ellipse_h5_path = tuple(path for path in eye_s3_paths if path.stem == "ellipses")
    if not ellipse_h5_path:
        raise FileNotFoundError(
            f"No ellipse h5 file found for {session}. Check dlc eye capsule"
        )

    # verbatim from allensdk: allensdk.brain_observatory.behavior.eye_tracking_processing.py
    eye_tracking_fields = ["cr", "eye", "pupil"]

    eye_tracking_dfs = []
    for field_name in eye_tracking_fields:
        df_ellipse = h5_to_dataframe(ellipse_h5_path[0], key_name=field_name)
        new_col_name_map = {
            col_name: f"{field_name}_{col_name}" for col_name in df_ellipse.columns
        }
        df_ellipse.rename(new_col_name_map, axis=1, inplace=True)
        eye_tracking_dfs.append(df_ellipse)

    eye_tracking_data = pd.concat(eye_tracking_dfs, axis=1)
    eye_tracking_data.index.name = "frame"

    # Values in the hdf5 may be complex (likely an artifact of the ellipse
    # fitting process). Take only the real component.
    eye_tracking_data = eye_tracking_data.apply(lambda x: np.real(x.to_numpy()))

    return eye_tracking_data.astype(float)


def get_dlc_session_model_dataframe_from_h5(
    session: str, model_name: str
) -> pd.DataFrame:
    """
    >>> df_model = get_dlc_session_model_dataframe_from_h5('676909_2023-12-12', 'dlc_eye')
    >>> len(df_model)
    512347
    """
    model_s3_paths = MODEL_FUNCTION_MAPPING[model_name](session)
    # TODO figure out ellipses.h5
    h5_files = tuple(path for path in model_s3_paths if path.suffix == ".h5")

    if not h5_files:
        raise FileNotFoundError(
            f"No h5 files found for {session}. Check {model_name} capsule"
        )

    df_model = h5_to_dataframe(h5_files[0])
    return df_model


def get_pose_series_from_dataframe(
    session: str, df: pd.DataFrame, video_timestamps: npt.NDArray[np.float64]
) -> list[ndx_pose.pose.PoseEstimationSeries]:
    # https://github.com/DeepLabCut/DLC2NWB/blob/main/dlc2nwb/utils.py#L189
    pose_estimations_series = []
    for keypoint, xy_positions in df.groupby(level="bodyparts", axis=1, sort=False):
        data = xy_positions.to_numpy()
        pose_estimation_series = ndx_pose.pose.PoseEstimationSeries(
            name=f"{session}_{keypoint}",
            description=f"Keypoint {keypoint} from session {session}.",
            data=data[:, :2],
            unit="pixels",
            reference_frame="(0,0) corresponds to the bottom left corner of the video.",
            timestamps=video_timestamps,
            confidence=data[:, 2],
            confidence_definition="Softmax output of the deep neural network.",
        )
        pose_estimations_series.append(pose_estimation_series)

    return pose_estimations_series


if __name__ == "__main__":
    import doctest

    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )
